@@ -3,11 +3,64 @@ IMGO - Compile, process, and augment image data.
 -------------------------------------------------
 AUGTOOLS module: 
 
-Last updated: version 1.0.2
+Last updated: version 2.0.0
 
 Classes
 -------        
 Augmenter: Class representing an image augmentation system.
+
+    Class Attributes:
+        rotate_range (list or tuple): range in degrees from which a random 
+        rotation angle is drawn.
+        -
+        shear_range (list or tuple): range in degrees from which a random 
+        shear angle is drawn.
+        -
+        dropout_pair (list or tuple): amount and degree of random pixel 
+        dropout applied.
+        -
+        x_shift (list or tuple): range relative to image size from which 
+        a random x-axis translation is drawn.
+        -
+        y_shift (list or tuple): range relative to image size from which 
+        a random y-axis translation is drawn.
+        -
+        clip_limit (list or tuple): range from which random clip limits 
+        for CLAH equalization are drawn.
+        -
+        pwa_scale (list or tuple): range relative to image size from which 
+        random piecewise affine translations are drawn.
+        -
+        h_flip (float): proportion of images on which a horizontal flip is 
+        performed.
+        -
+        v_flip (float): proportion of images on which a vertical flip is 
+        performed.
+        -
+        g_sev (int): amount (or range if batch is randomized) of severity 
+        with which gaussian noise is applied.
+        -
+        b_sev (int): amount (or range if batch is randomized) of severity 
+        with which brightness is altered.
+        -
+        e_sev (int): amount (or range if batch is randomized) of severity 
+        with which elastic transformation is applied.
+        -
+        contrast (bool): whether or not random contrast adjustment is 
+        applied.
+        -
+        sharpness (bool): whether or not random sharpness adjustment is 
+        applied.
+        -
+        randomize_batch (bool): whether the augmentation parameters 
+        applied are fixed as the input given or are randomly drawn from a 
+        range inferred from the input given.
+        -
+        f_list (list): list of augmentation functions that have been 
+        applied.
+        -
+        argno (int): number of augmentation functions that have been 
+        applied.   
     
     Class Methods:
         init: constructs all the necessary attributes for the augmenter.
@@ -48,8 +101,8 @@ Augmenter: Class representing an image augmentation system.
         -
         display_sample: randomly augment and display image.
         -
-        augment_flow: apply (and save) augmentation to images located 
-        in directories on disk.
+        augment_flow: apply augmentation to images located in directories 
+        on disk, and save augmented images to disk.
 """
 
 import os
@@ -59,11 +112,11 @@ import random
 import imageio
 import cv2
 import matplotlib.pyplot as plt
+import h5py
 from imgaug import augmenters as iaa
 from tqdm import tqdm
 from sklearn.model_selection import train_test_split
 from send2trash import send2trash
-from datetime import datetime
 
 # ------------------------------------------------------------------------
 
@@ -81,65 +134,60 @@ class Augmenter:
 
     Attributes
     ----------
-    rotate_range (list or tuple) optional/dependent on init: range in
-    degrees from which a random rotation angle is drawn.
+    rotate_range (list or tuple): range in degrees from which a random
+    rotation angle is drawn.
     -
-    shear_range (list or tuple) optional/dependent on init: range in
-    degrees from which a random shear angle is drawn.
+    shear_range (list or tuple): range in degrees from which a random
+    shear angle is drawn.
     -
-    dropout_pair (list or tuple) optional/dependent on init: amount and
-    degree of random pixel dropout applied.
+    dropout_pair (list or tuple): amount and degree of random pixel
+    dropout applied.
     -
-    x_shift (list or tuple) optional/dependent on init: range relative
-    to image size from which a random x-axis translation is drawn.
+    x_shift (list or tuple): range relative to image size from which a
+    random x-axis translation is drawn.
     -
-    y_shift (list or tuple) optional/dependent on init: range relative
-    to image size from which a random y-axis translation is drawn.
+    y_shift (list or tuple): range relative to image size from which a
+    random y-axis translation is drawn.
     -
-    clip_limit (list or tuple) optional/dependent on init: range from
-    which random clip limits for CLAH equalization are drawn.
+    clip_limit (list or tuple): range from which random clip limits for
+    CLAH equalization are drawn.
     -
-    pwa_scale (list or tuple) optional/dependent on init: range relative
-    to image size from which random piecewise affine translations are
-    drawn.
+    pwa_scale (list or tuple): range relative to image size from which
+    random piecewise affine translations are drawn.
     -
-    h_flip (float) optional/dependent on init: proportion of images on
-    which a horizontal flip is performed.
+    h_flip (float): proportion of images on which a horizontal flip is
+    performed.
     -
-    v_flip (float) optional/dependent on init: proportion of images on
-    which a vertical flip is performed.
+    v_flip (float): proportion of images on which a vertical flip is
+    performed.
     -
-    g_sev (int) optional/dependent on init: amount (or range if batch
-    is randomized) of severity with which gaussian noise is applied.
+    g_sev (int): amount (or range if batch is randomized) of severity with
+    which gaussian noise is applied.
     -
-    b_sev (int) optional/dependent on init: amount (or range if batch
-    is randomized) of severity with which brightness is altered.
+    b_sev (int): amount (or range if batch is randomized) of severity with
+    which brightness is altered.
     -
-    e_sev (int) optional/dependent on init: amount (or range if batch
-    is randomized) of severity with which elastic transformation is
+    e_sev (int): amount (or range if batch is randomized) of severity with
+    which elastic transformation is applied.
+    -
+    contrast (bool): whether or not random contrast adjustment is applied.
+    -
+    sharpness (bool): whether or not random sharpness adjustment is
     applied.
     -
-    contrast (bool) optional/dependent on init: whether or not random
-    contrast adjustment is applied.
+    randomize_batch (bool): whether the augmentation parameters applied
+    are fixed as the input given or are randomly drawn from a range
+    inferred from the input given.
     -
-    sharpness (bool) optional/dependent on init: whether or not random
-    sharpness adjustment is applied.
+    f_list (list): list of augmentation functions that have been applied.
     -
-    randomize_batch (bool) optional/dependent on init: whether the
-    augmentation parameters applied are fixed as the input given or
-    are randomly drawn from a range inferred from the input given.
-    -
-    f_list (list) optional/dependent on init: list of augmentation
-    functions that have been applied.
-    -
-    argno (int) optional/dependent on init: number of augmentation
-    functions that have been applied.
+    argno (int): number of augmentation functions that have been applied.
 
     Methods
     -------
     init: constructs all the necessary attributes for the augmenter.
     -
-    details: display summary information about the augmenter.
+    summary: display summary information about the augmenter.
     -
     aug_rotate: apply imgaug Affine rotation to image.
     -
@@ -169,12 +217,14 @@ class Augmenter:
     -
     aug_sharpness: apply imgaug EnhanceSharpness to image.
     -
-    simple_augment: apply one or more augmentations to image.
+    simple_augment: apply one or more augmentations to an image.
     -
     random_augment: randomly apply one or more augmentations to image.
     -
     display_sample: randomly augment and display image.
-
+    -
+    augment_flow: apply augmentation to images located in directories
+    on disk, and save augmented images to disk.
     """
 
     def __init__(
@@ -398,16 +448,16 @@ class Augmenter:
 
     #     ----------
 
-    def aug_rotate(self, img, prenorm=False):
+    def aug_rotate(self, img, pre_norm=False):
 
         """
         Rotates the image by a random angle within the range given.
-        Note: Set prenorm to True if image has been normalized to [0,1].
+        Note: Set pre_norm to True if image has been normalized to [0,1].
         """
 
         if self.rotate_range is not None:
             rotate = iaa.Affine(rotate=self.rotate_range, mode="edge")
-            if prenorm:
+            if pre_norm:
                 img = (img * 255).astype(np.uint8)
                 rotated_img = (rotate(image=img)) / 255
             else:
@@ -418,17 +468,17 @@ class Augmenter:
 
     #     ----------
 
-    def aug_shear(self, img, prenorm=False):
+    def aug_shear(self, img, pre_norm=False):
 
         """
         Shear-transforms the image by a random angle within the range
         given.
-        Note: Set prenorm to True if image has been normalized to [0,1].
+        Note: Set pre_norm to True if image has been normalized to [0,1].
         """
 
         if self.shear_range is not None:
             shear = iaa.Affine(shear=self.shear_range, mode="edge")
-            if prenorm:
+            if pre_norm:
                 img = (img * 255).astype(np.uint8)
                 shear_img = (shear(image=img)) / 255
             else:
@@ -439,11 +489,11 @@ class Augmenter:
 
     #     ----------
 
-    def aug_dropout(self, img, prenorm=False):
+    def aug_dropout(self, img, pre_norm=False):
 
         """
         Drops a portion of pixel values to zero.
-        Note: Set prenorm to True if image has been normalized to [0,1].
+        Note: Set pre_norm to True if image has been normalized to [0,1].
         """
 
         if self.dropout_pair is not None:
@@ -462,7 +512,7 @@ class Augmenter:
                     self.dropout_pair[0],
                     size_percent=self.dropout_pair[1],
                 )
-            if prenorm:
+            if pre_norm:
                 img = (img * 255).astype(np.uint8)
                 dropout_img = (dropout(image=img)) / 255
             else:
@@ -473,17 +523,17 @@ class Augmenter:
 
     #     ----------
 
-    def aug_x_shift(self, img, prenorm=False):
+    def aug_x_shift(self, img, pre_norm=False):
 
         """
         Shifts the image on the x-axis by a random amount within the
         range given.
-        Note: Set prenorm to True if image has been normalized to [0,1].
+        Note: Set pre_norm to True if image has been normalized to [0,1].
         """
 
         if self.x_shift is not None:
             x_shift = iaa.TranslateX(percent=self.x_shift, mode="edge")
-            if prenorm:
+            if pre_norm:
                 img = (img * 255).astype(np.uint8)
                 x_shift_img = (x_shift(image=img)) / 255
             else:
@@ -494,17 +544,17 @@ class Augmenter:
 
     #     ----------
 
-    def aug_y_shift(self, img, prenorm=False):
+    def aug_y_shift(self, img, pre_norm=False):
 
         """
         Shifts the image on the y-axis by a random amount within the
         range given.
-        Note: Set prenorm to True if image has been normalized to [0,1].
+        Note: Set pre_norm to True if image has been normalized to [0,1].
         """
 
         if self.y_shift is not None:
             y_shift = iaa.TranslateY(percent=self.y_shift, mode="edge")
-            if prenorm:
+            if pre_norm:
                 img = (img * 255).astype(np.uint8)
                 y_shift_img = (y_shift(image=img)) / 255
             else:
@@ -515,17 +565,17 @@ class Augmenter:
 
     #     ----------
 
-    def aug_clahe(self, img, prenorm=False):
+    def aug_clahe(self, img, pre_norm=False):
 
         """
         Applies a CLAH equalization using a random clip limit within
         the range given.
-        Note: Set prenorm to True if image has been normalized to [0,1].
+        Note: Set pre_norm to True if image has been normalized to [0,1].
         """
 
         if self.clip_limit is not None:
             clahe = iaa.CLAHE(clip_limit=self.clip_limit)
-            if prenorm:
+            if pre_norm:
                 img = (img * 255).astype(np.uint8)
                 clahe_img = (clahe(image=img)) / 255
             else:
@@ -536,17 +586,17 @@ class Augmenter:
 
     #     ----------
 
-    def aug_pwa(self, img, prenorm=False):
+    def aug_pwa(self, img, pre_norm=False):
 
         """
         Shifts local pixel areas by a random amount within the range
         given.
-        Note: Set prenorm to True if image has been normalized to [0,1].
+        Note: Set pre_norm to True if image has been normalized to [0,1].
         """
 
         if self.pwa_scale is not None:
             pwa = iaa.PiecewiseAffine(scale=self.pwa_scale, mode="edge")
-            if prenorm:
+            if pre_norm:
                 img = (img * 255).astype(np.uint8)
                 pwa_img = (pwa(image=img)) / 255
             else:
@@ -557,16 +607,16 @@ class Augmenter:
 
     #     ----------
 
-    def aug_h_flip(self, img, prenorm=False):
+    def aug_h_flip(self, img, pre_norm=False):
 
         """
         Flips a random subset of images horizontally.
-        Note: Set prenorm to True if image has been normalized to [0,1].
+        Note: Set pre_norm to True if image has been normalized to [0,1].
         """
 
         if self.h_flip is not None:
             h_flip = iaa.Fliplr(self.h_flip)
-            if prenorm:
+            if pre_norm:
                 img = (img * 255).astype(np.uint8)
                 h_flip_img = (h_flip(image=img)) / 255
             else:
@@ -577,16 +627,16 @@ class Augmenter:
 
     #     ----------
 
-    def aug_v_flip(self, img, prenorm=False):
+    def aug_v_flip(self, img, pre_norm=False):
 
         """
         Flipts a random subset of images vertically.
-        Note: Set prenorm to True if image has been normalized to [0,1].
+        Note: Set pre_norm to True if image has been normalized to [0,1].
         """
 
         if self.v_flip is not None:
             v_flip = iaa.Flipud(self.v_flip)
-            if prenorm:
+            if pre_norm:
                 img = (img * 255).astype(np.uint8)
                 v_flip_img = (v_flip(image=img)) / 255
             else:
@@ -597,11 +647,11 @@ class Augmenter:
 
     #     ----------
 
-    def aug_g_noise(self, img, prenorm=False):
+    def aug_g_noise(self, img, pre_norm=False):
 
         """
         Adds Gaussian noise to the image.
-        Note: Set prenorm to True if image has been normalized to [0,1].
+        Note: Set pre_norm to True if image has been normalized to [0,1].
         """
 
         if self.g_sev is not None:
@@ -620,7 +670,7 @@ class Augmenter:
             if g_noise is None:
                 return img
             else:
-                if prenorm:
+                if pre_norm:
                     img = (img * 255).astype(np.uint8)
                     g_noise_img = (g_noise(image=img)) / 255
                 else:
@@ -631,11 +681,11 @@ class Augmenter:
 
     #     ----------
 
-    def aug_brightness(self, img, prenorm=False):
+    def aug_brightness(self, img, pre_norm=False):
 
         """
         Adjusts brightness of the image.
-        Note: Set prenorm to True if image has been normalized to [0,1].
+        Note: Set pre_norm to True if image has been normalized to [0,1].
         """
 
         if self.b_sev is not None:
@@ -654,7 +704,7 @@ class Augmenter:
             if brightness is None:
                 return img
             else:
-                if prenorm:
+                if pre_norm:
                     img = (img * 255).astype(np.uint8)
                     brightness_img = (brightness(image=img)) / 255
                 else:
@@ -665,11 +715,11 @@ class Augmenter:
 
     #     ----------
 
-    def aug_elastic(self, img, prenorm=False):
+    def aug_elastic(self, img, pre_norm=False):
 
         """
         Applies elastic transformation to the image.
-        Note: Set prenorm to True if image has been normalized to [0,1].
+        Note: Set pre_norm to True if image has been normalized to [0,1].
         """
 
         if self.e_sev is not None:
@@ -688,7 +738,7 @@ class Augmenter:
             if elastic is None:
                 return img
             else:
-                if prenorm:
+                if pre_norm:
                     img = (img * 255).astype(np.uint8)
                     elastic_img = (elastic(image=img)) / 255
                 else:
@@ -699,30 +749,30 @@ class Augmenter:
 
     #     ----------
 
-    def aug_contrast(self, img, prenorm=False):
+    def aug_contrast(self, img, pre_norm=False):
 
         """
         Adusts contrast to a random subset of images.
-        Note: Set prenorm to True if image has been normalized to [0,1].
+        Note: Set pre_norm to True if image has been normalized to [0,1].
         """
 
         if self.contrast == True:
             contrast = iaa.pillike.EnhanceContrast()
             if self.randomize_batch == True:
                 if np.random.randint(2) == True:
-                    if prenorm:
+                    if pre_norm:
                         img = (img * 255).astype(np.uint8)
                         contrast_img = (contrast(image=img)) / 255
                     else:
                         contrast_img = contrast(image=img)
                 else:
-                    if prenorm:
+                    if pre_norm:
                         img = (img * 255).astype(np.uint8)
                         contrast_img = (contrast(image=img)) / 255
                     else:
                         contrast_img = img
             else:
-                if prenorm:
+                if pre_norm:
                     img = (img * 255).astype(np.uint8)
                     contrast_img = (contrast(image=img)) / 255
                 else:
@@ -733,30 +783,30 @@ class Augmenter:
 
     #     ----------
 
-    def aug_sharpness(self, img, prenorm=False):
+    def aug_sharpness(self, img, pre_norm=False):
 
         """
         Adjusts sharpness to a random subset of images.
-        Note: Set prenorm to True if image has been normalized to [0,1].
+        Note: Set pre_norm to True if image has been normalized to [0,1].
         """
 
         if self.sharpness == True:
             sharpness = iaa.pillike.EnhanceSharpness()
             if self.randomize_batch == True:
                 if np.random.randint(2) == True:
-                    if prenorm:
+                    if pre_norm:
                         img = (img * 255).astype(np.uint8)
                         sharpness_img = (sharpness(image=img)) / 255
                     else:
                         sharpness_img = sharpness(image=img)
                 else:
-                    if prenorm:
+                    if pre_norm:
                         img = (img * 255).astype(np.uint8)
                         sharpness_img = (sharpness(image=img)) / 255
                     else:
                         sharpness_img = img
             else:
-                if prenorm:
+                if pre_norm:
                     img = (img * 255).astype(np.uint8)
                     sharpness_img = (sharpness(image=img)) / 255
                 else:
@@ -767,42 +817,51 @@ class Augmenter:
 
     #     ----------
 
-    def simple_augment(self, img, prenorm=False):
+    def simple_augment(self, img, pre_norm=False):
 
         """
         Performs the augmentations (for which a parameter was provided)
         in order.
-        Note: Set prenorm to True if image has been normalized to [0,1].
+        Note: Set pre_norm to True if image has been normalized to [0,1].
         """
-        if prenorm:
+        if pre_norm:
             img = (img * 255).astype(np.uint8)
-        order = list(range(self.argno))
+        order = np.arange(self.argno)
         for i in order:
             function = self.f_list[i]
             outp = function(img)
             img = outp
-        if prenorm:
+        if pre_norm:
             return outp / 255
         else:
             return outp
 
     #     ----------
 
-    def random_augment(self, img, prenorm=False):
+    def random_augment(self, img, full_set=True, pre_norm=False):
 
         """
         Performs the augmentations (for which a parameter was provided)
         in a random order.
-        Note: Set prenorm to True if image has been normalized to [0,1].
+        Note: Set pre_norm to True if image has been normalized to [0,1].
         """
-        if prenorm:
+        if pre_norm:
             img = (img * 255).astype(np.uint8)
-        order = random.sample(list(range(self.argno)), self.argno)
+        if full_set:
+            order = np.random.choice(
+                np.arange(self.argno), size=self.argno, replace=False
+            )
+        else:
+            order = np.random.choice(
+                np.arange(self.argno),
+                size=np.random.randint(self.argno),
+                replace=False,
+            )
         for i in order:
             function = self.f_list[i]
             outp = function(img)
             img = outp
-        if prenorm:
+        if pre_norm:
             return outp / 255
         else:
             return outp
@@ -816,7 +875,7 @@ class Augmenter:
         n_rows,
         n_cols,
         augment_type="random",
-        prenorm=False,
+        pre_norm=False,
     ):
 
         """
@@ -857,7 +916,7 @@ class Augmenter:
             for the display. If "simple", the simple_augment method
             will be used. Defaults to "random".
             -
-            prenorm (bool) optional: whether or not the image data
+            pre_norm (bool) optional: whether or not the image data
             has been normalized to the range [0,1]. If it has, not
             setting this to True will result in error. Defaults to
             False.
@@ -875,17 +934,21 @@ class Augmenter:
             img = imageio.imread(sample.iloc[0]["image"])
 
         elif source_type == "ids":
-            img = source.X_data[
-                np.random.randint(source.X_data.shape[0])
-            ]
+            if source.shadow["data"][0] is None:
+                img = source.shadow["train"][0][
+                    np.random.randint(
+                        source.shadow["train"][0].shape[0]
+                    )
+                ]
+            else:
+                img = source.shadow["data"][0][
+                    np.random.randint(source.shadow["data"][0].shape[0])
+                ]
         else:
             raise Exception(
                 "Must choose valid source type: either 'path' or 'ids'."
             )
 
-        if prenorm:
-            img = (img * 255).astype(np.uint8)
-
         plt.rcParams["font.family"] = "sans-serif"
         plt.rcParams["font.sans-serif"] = "Helvetica"
         plt.rcParams["text.color"] = "#333F4B"
@@ -899,8 +962,6 @@ class Augmenter:
         for i in tqdm(range(2, (n_rows * n_cols) + 1)):
             if augment_type == "simple":
                 aug_img = self.simple_augment(img)
-            elif augment_type == "random":
-                aug_img = self.random_augment(img)
             else:
                 aug_img = self.random_augment(img)
             ax = fig.add_subplot(n_rows, n_cols, i)
@@ -988,8 +1049,7 @@ class Augmenter:
         if class_selection:
             if type(class_selection) is not list:
                 raise Exception(
-                    f"Class selection must be a list;"
-                    + f" {type(class_selection)} given."
+                    f"Class selection must be a list; {type(class_selection)} given."
                 )
             else:
                 df = df.loc[df["class"].isin(class_selection)]
@@ -1021,8 +1081,7 @@ class Augmenter:
             for a in all_class_arrays:
                 if size - a.shape[0] < 0:
                     raise Exception(
-                        "Cannot accept size argument smaller"
-                        + " than number of images."
+                        "Cannot accept size argument smaller than number of images."
                     )
                 else:
                     to_augment = np.random.randint(
@@ -1037,10 +1096,19 @@ class Augmenter:
 
         if save:
             if no_class == True:
-                if os.path.exists(f"{base_path}/augmented_images"):
-                    pass
-                else:
-                    os.mkdir(f"{base_path}/augmented_images")
+                my_path = "imgo_output/augtools/augment_flow"
+
+                r = None
+                for i in my_path.split("/"):
+                    if r == None:
+                        if not os.path.exists(i):
+                            os.mkdir(i)
+                        r = i
+                    else:
+                        if not os.path.exists(r + "/" + i):
+                            os.mkdir(r + "/" + i)
+                        r = r + "/" + i
+
                 for i, j in tqdm(
                     enumerate(all_aug_indices[0]),
                     total=len(all_aug_indices[0]),
@@ -1048,16 +1116,10 @@ class Augmenter:
                     aug_img = self.random_augment(
                         all_class_arrays[0][j]
                     )
-                    imageio.imwrite(
-                        f"{base_path}/augmented_images/"
-                        + f"aug_{i+1}.jpg",
-                        aug_img,
-                    )
+                    imageio.imwrite(f"{r}/aug_{i+1}.jpg", aug_img)
                     aug_imgs.append(aug_img)
-                print(
-                    f"Augmented images saved in {base_path}/"
-                    + "augmented_images/"
-                )
+                print("Augmented images saved successfully.")
+
             else:
                 for x in range(len(all_class_arrays)):
                     for i, j in tqdm(
@@ -1068,14 +1130,12 @@ class Augmenter:
                             all_class_arrays[x][j]
                         )
                         imageio.imwrite(
-                            f"{base_path}/{class_list[x]}"
-                            + f"/{class_list[x]}_aug_{i+1}.jpg",
+                            f"{base_path}/{class_list[x]}/{class_list[x]}_aug_{i+1}.jpg",
                             aug_img,
                         )
                         aug_imgs.append(aug_img)
                 print(
-                    "Augmented images saved in class subdirectories"
-                    + f" within {base_path}/."
+                    "Augmented images saved successfully in class subdirectories."
                 )
         else:
             if no_class == True:
@@ -1089,7 +1149,6 @@ class Augmenter:
                     aug_imgs.append(aug_img)
 
             else:
-
                 for x in range(len(all_class_arrays)):
                     for i, j in tqdm(
                         enumerate(all_aug_indices[x]),
